import mne
from mne.io.brainvision.brainvision import RawBrainVision
from os.path import dirname, basename, exists, join
import numpy as np
from scipy.io import loadmat
from mne import Epochs

class RawCLAM(RawBrainVision):

    """Initialize a RawCLAM object.

    Parameters:
    -----------
    path : str
        The path to the BrainVision header file (.vhdr).
    l_freq_target : float
        Lower edge of the target frequency range.
    h_freq_target : float
        Higher edge of the target frequency range.
    tmin : float or None, optional
        For 'trial_wise' designs, this is the start time of the trial relative to the target phase marker.
        Required paramter when trigger markers are provided (e.g. marker_definition != {}).
    tmax : float or None, optional
        For 'trial_wise' designs, this is the end time of the trial relative to the target phase marker.
        Required paramter when trigger markers are provided (e.g. marker_definition != {}).
    n_chs : int, optional
        Number of EEG channels in the data (including bads).
    design : str, optional
        The experimental design type.
        'trial_wise' means that multiple phase lags were tested in the session.
        'session_wise' means that a single phase lag was tested in the session (e.g. patient treatment).
    misc_channels : list, optional
        List of miscellaneous channel names (not EEG or ECG).
    marker_definition : dict, optional
        Dictionary containing marker definitions.
        Mapping from target phase markers (e.g. 1 - 6) to target phases [-pi, pi].
        Target phases can also take the values 'ns' (no stimulation) or 'ol' (open-loop stimulation).
    sfreq : float or None, optional
        New sampling frequency, or None if the data should not be resampled.
    ignore_calibration_files: bool, optional
        If True, the user will be prompted to select bad channels and target spatial component regardless
        of the presence of calibration (.mat) files in the data folder
    default_bads : list of str, optional
        The user may specify the list of channels that are marked bad by default in viz.set_bads

    Notes:
    ------
    This class extends RawBrainVision and provides additional functionality specific to CLAM-NIBS experiments.

    Raises:
    -------
    Exception:
        - If data with CLAM-NIBS is loaded but no forward model is present in the data folder.
        - If data with CLAM-NIBS is loaded but no dipole sign flip is present in the data folder.
    """

    def __init__(
            self,
            path,
            l_freq_target,
            h_freq_target,
            tmin = None,
            tmax = None,
            n_chs=64,
            design='trial_wise',
            ecg_channels=[],
            misc_channels=['envelope',
                           'envelope_am'],
            marker_definition={1: (0 / 6) * 2 * np.pi,
                               2: (1 / 6) * 2 * np.pi,
                               3: (2 / 6) * 2 * np.pi,
                               4: (3 / 6) * 2 * np.pi,
                               5: (4 / 6) * 2 * np.pi,
                               6: (5 / 6) * 2 * np.pi},
            sfreq=None,
            ignore_calibration_files = False,
            default_bads=['Fp1', 'Fpz', 'Fp2', 'F9', 'FT9', 'TP9', 'F10', 'FT10', 'TP10']):
        super().__init__(path, preload=True)
        self.n_chs = n_chs
        if sfreq is not None:
            self.resample(sfreq)
        self.filter(l_freq_target, h_freq_target, picks=['envelope'])
        folder_path = dirname(path)
        misc_channels = [ch for ch in misc_channels if ch in self.ch_names]
        if 'no_stim' in path.lower():
            self.is_stim = False
        else:
            self.is_stim = True
        self.design = design
        self.set_channel_types({
            **{ch: 'ecg' for ch in ecg_channels},
            **{ch: 'misc' for ch in misc_channels}})
        self.l_freq_target = l_freq_target
        self.h_freq_target = h_freq_target
        
        if marker_definition:
            if tmin is None or tmax is None:
                raise Exception(
                    'tmin and tmax are required parameters (cannot be None) for epoching data when a marker definition is provided')
        
        self.marker_definition = marker_definition
        self.tmin = tmin
        self.tmax = tmax
        self.set_montage('easycap-M1', match_case=False, on_missing='warn')
        if design == 'trial_wise':
            self.participant = basename(dirname(path))
            self.session = 'T01'
        else:
            self.participant = basename(dirname(dirname(path)))
            self.session = basename(dirname(path))
        
        exclude_idx_file_path = join(folder_path, 'exclude_idx.mat')
        if exists(exclude_idx_file_path) and not ignore_calibration_files:
            exclude_idx_mat = loadmat(exclude_idx_file_path)['exclude_idx']
            if len(exclude_idx_mat) == 0:
                bads = np.array([])
            else:
                bads = np.array(self.ch_names)[exclude_idx_mat[0] - 1]
            self.info['bads'] = list(bads)
        else:
            from . import viz
            viz.set_bads(self, default_bads)
            
        p_target_file_path = join(folder_path, 'P_TARGET_{:d}.mat'.format(int(n_chs)))
        if exists(p_target_file_path) and not ignore_calibration_files:
            self.forward_full = loadmat(p_target_file_path)['P_TARGET_{:d}'.format(int(n_chs))][0]
        else:
            if self.is_stim:
                raise Exception(
                    'Data with CLAM-tACS was loaded, but no forward model was present in the data folder')
            
            from . import beamformer
            beamformer.set_forward(self, 1, 30)
            
        flip_file_path = join(folder_path, 'flip.mat')
        
        if exists(flip_file_path) and not ignore_calibration_files:
            self.flip = loadmat(flip_file_path)['flip'][0]
        else:
            if self.is_stim:
                raise Exception(
                    'Data with CLAM-tACS was loaded, but no dipole sign flip was present in the data folder')
            
            from . import beamformer
            beamformer.set_flip(self)
        


class EpochsCLAM(Epochs):
    
    """Initialize an EpochsCLAM object.

    Parameters:
    -----------
    raw : RawCLAM object
        The EEG data.
    apply_hil : bool
        Whether to apply Hilbert transformation to the EEG data to obtain the analytic signal.

    Notes:
    ------
    This class extends Epochs and provides additional functionality specific to CLAM-NIBS experiments.

    Attributes:
    -----------
    design : str
        The experimental design type ('trial_wise' or 'session_wise').
    l_freq_target : float
        Lower edge of the target frequency range (in Hz).
    h_freq_target : float
        Higher edge of the target frequency range (in Hz).
    marker_definition : dict
        Dictionary containing marker definitions.
        Mapping from target phase markers (e.g. 1 - 6) to target phases [-pi, pi].
    is_stim : bool
        Indicates whether the data was recorded in the presence of CLAM-NIBS or not.
    participant : str
        Participant identifier.
    session : str
        Session identifier.
    forward_full : ndarray or None
        Forward model for all EEG channels (with zero for bads).
    flip : integer
        Sign flip for dipole (-1 or 1).

    Raises:
    -------
    None
    """
    
    def __init__(self, raw, apply_hil=True):
        if not isinstance(raw, RawCLAM):
            raise Exception(
                'Please use get_raw to create a RawCLAM object before using get_epochs to create an EpochsCLAM object')
        target_codes = list(raw.marker_definition.keys())
        events = mne.events_from_annotations(raw)[0]
        tmin = raw.tmin
        tmax = raw.tmax
        if apply_hil:
            picks = [ch_name for ch_name, ch_type in zip(raw.ch_names, raw.get_channel_types()) if ch_type=='eeg']+['envelope']
            raw_out = raw.copy().apply_hilbert(picks=picks, envelope=False)
        else:
            raw_out = raw
        super().__init__(
            raw_out,
            events,
            event_id=target_codes,
            on_missing='ignore',
            tmin=tmin,
            tmax=tmax,
            baseline=None,
            proj=False,
            preload=True)
        self.design = raw.design
        self.l_freq_target = raw.l_freq_target
        self.h_freq_target = raw.h_freq_target
        self.marker_definition = raw.marker_definition
        self.is_stim = raw.is_stim
        self.participant = raw.participant
        self.session = raw.session
        self.forward_full = raw.forward_full
        self.flip = raw.flip
        self.n_chs = raw.n_chs


import numpy as np
from mne.utils.check import _check_sphere
from mne.viz.topomap import _get_pos_outlines, _draw_outlines
from matplotlib import rcParams
import matplotlib.pyplot as plt
from functools import partial
from .base import RawCLAM, EpochsCLAM

def _onpick_sensor(event, fig, ax, pos, ch_names, bads, scatter):
    if event is not None:
        if event.mouseevent.inaxes != ax:
            return
        ind = event.ind[0]
        ch_name = ch_names[ind]
        if ch_name in bads:
            bads.remove(ch_name)
        else:
            bads.append(ch_name)
            
    edgecolors = ['r' if ch in bads else 'k' for ch in ch_names]
    scatter.set_edgecolors(edgecolors)
    fig.canvas.draw()


def set_bads(obj, default_bads):
    
    """Interactive tool to mark bad channels in EEG data.

    This function provides an interactive visualization to mark bad channels in EEG data.
    It displays a scalp plot with the channels labeled, allowing the user to click on
    channels to mark them as bad.

    Parameters:
    -----------
    obj : RawCLAM or EpochsCLAM object
        The RawCLAM or EpochsCLAM object containing EEG data.
    default_bads: list of str
        The channels that are marked bad by default

    Raises:
    -------
    Exception:
        If the input object is not an instance of RawCLAM or EpochsCLAM.

    Notes:
    ------
    - This function requires Matplotlib to display the interactive visualization.
    - Bad channels are marked by clicking on the corresponding channel in the plot.
    - The plot window must be closed to finalize the selection of bad channels.
    - The updated list of bad channels is stored in the `bads` attribute of the input object.
    """
    
    if not (isinstance(obj, RawCLAM) or isinstance(obj, EpochsCLAM)):
        raise Exception('set_bads can only be applied to RawCLAM or EpochsCLAM objects')

    info = obj.info
    ch_names = obj.ch_names

    pos = np.empty((len(info["chs"]), 3))
    for ci, ch in enumerate(info['chs']):
        pos[ci] = ch["loc"][:3]

    sphere = _check_sphere(None, info)

    subplot_kw = dict()
    fig, ax = plt.subplots(1, figsize=(
        max(rcParams["figure.figsize"]),) * 2, subplot_kw=subplot_kw)

    ax.text(0, 0, "", zorder=1)

    pos, outlines = _get_pos_outlines(
        info, range(len(ch_names)), sphere, to_sphere=True)
    _draw_outlines(ax, outlines)
    # DRAW SERIES OF LINES HERE FOR CONNECTIVITY GRAPH
    scatter = ax.scatter(
        pos[:, 0],
        pos[:, 1],
        picker=True,
        clip_on=False,
        c='k',
        edgecolors='k',
        s=150,
        lw=2,
    )

    ax.set(aspect="equal")
    fig.subplots_adjust(left=0, bottom=0, right=1, top=1)
    ax.axis("off")

    indices = range(len(pos))
    for idx in indices:
        this_pos = pos[idx]
        ax.text(
            this_pos[0],
            this_pos[1] - 0.007,
            ch_names[idx],
            ha="center",
            va="center",
        )
    xmin, ymin, xmax, ymax = fig.get_window_extent().bounds
    renderer = fig.canvas.get_renderer()
    extents = [x.get_window_extent(renderer=renderer) for x in ax.texts]
    xmaxs = np.array([x.max[0] for x in extents])
    bad_xmax_ixs = np.nonzero(xmaxs > xmax)[0]
    if len(bad_xmax_ixs):
        needed_space = (xmaxs[bad_xmax_ixs] - xmax).max() / xmax
        fig.subplots_adjust(right=1 - 1.1 * needed_space)

    bads = [ch for ch in default_bads if ch in ch_names]
    
    picker = partial(
        _onpick_sensor,
        fig=fig,
        ax=ax,
        pos=pos,
        ch_names=ch_names,
        bads=bads,
        scatter=scatter
    )
    fig.canvas.mpl_connect("pick_event", picker)
    picker(None) # call to the update function with a dummy event to force it to draw the initial bad channels
    fig.set_size_inches(10, 12)
    ax.text(
        0.05,
        0.95,
        'Please mark all bad channels by clicking on them.\nClose the window when done',
        transform=ax.transAxes,
        verticalalignment='top',
        horizontalalignment='center',
        fontsize=14,
        weight='bold')
    plt.tight_layout()
    plt.show()
    obj.info['bads'] = bads


import numpy as np
from scipy import linalg
import mne
from scipy.signal import hilbert
from . import misc
from mne.stats import permutation_cluster_test
from scipy.sparse import coo_matrix
import scipy
from mne.stats import ttest_ind_no_p
from pingouin import circ_corrcl
from mne.viz.topomap import _get_pos_outlines
from mne.utils.check import _check_sphere
import matplotlib.pyplot as plt
import seaborn as sns
from mne.viz import plot_sensors
import pandas as pd
from .misc import _get_ixs_goods, _get_main_target_phase
from .base import RawCLAM, EpochsCLAM


def _find_n_nulls(A, B, D, M):
    errors = []
    for n_nulls in range(A.shape[0]):
        DI = np.ones(M.shape[0])
        DI[:n_nulls] = 0
        DI = np.diag(DI)
        P = M.dot(DI).dot(D)
        errors.append(linalg.norm(B - P @ A @ P.T))
    return np.argmin(errors)


def clean_sensor_data(obj_no_stim, obj_stim):
    
    """
    Use Stimulation Artifact Source Separation (SASS) [1] to remove electric stimulation artifacts 
    from EEG sensor data recorded during CLAM-NIBS. Bad sensors are excluded from the procedure,
    and interpolated after application of SASS.

    Parameters:
    -----------
    obj_no_stim : RawCLAM or EpochsCLAM
        The RawCLAM or EpochsCLAM object containing the sensor data without CLAM-NIBS stimulation.
    obj_stim : RawCLAM or EpochsCLAM
        The RawCLAM or EpochsCLAM object containing the sensor data with CLAM-NIBS stimulation.

    Raises:
    -------
    Exception
        If the input data types are not RawCLAM or EpochsCLAM objects.
        If the data with and without CLAM-NIBS have different bandpass filter or target frequency range settings.
    
    References:
    -----------
    [1] Haslacher, David, et al. "Stimulation artifact source separation (SASS) for assessing electric brain oscillations 
    during transcranial alternating current stimulation (tACS)." Neuroimage 228 (2021): 117571.
    """
    
    if not (isinstance(obj_no_stim, RawCLAM) or isinstance(obj_no_stim, EpochsCLAM)) and \
            (isinstance(obj_stim, RawCLAM) or isinstance(obj_stim, EpochsCLAM)):
        raise Exception('clean_sensor_data can only be applied to RawCLAM or EpochsCLAM objects')

    equal_l_freq = obj_no_stim.info['highpass'] == obj_stim.info['highpass']
    equal_h_freq = obj_no_stim.info['lowpass'] == obj_stim.info['lowpass']
    equal_l_freq_target = obj_no_stim.l_freq_target == obj_stim.l_freq_target
    equal_h_freq_target = obj_no_stim.h_freq_target == obj_stim.h_freq_target
    if not (
            equal_l_freq and equal_h_freq and equal_l_freq_target and equal_h_freq_target):
        raise Exception(
            'Data with and without CLAM-tACS have different bandpass-filter or target frequency range settings')
    l_freq = obj_no_stim.info['highpass']
    h_freq = obj_no_stim.info['lowpass']
    l_freq_target = obj_no_stim.l_freq_target
    h_freq_target = obj_no_stim.h_freq_target
    ixs_goods = _get_ixs_goods(obj_no_stim)

    if not (l_freq == l_freq_target and h_freq == h_freq_target):
        raise Exception(
            'Data must be filtered into the target frequency range to apply SASS')
        
    if isinstance(obj_stim, RawCLAM):
        A = np.cov(obj_stim.get_data(ixs_goods))
    else:
        A = np.cov(
            np.concatenate(
                obj_stim.get_data(ixs_goods),
                axis=-1))
        
    if isinstance(obj_no_stim, RawCLAM):
        B = np.cov(obj_no_stim.get_data(ixs_goods))
    else:
        B = np.cov(
            np.concatenate(
                obj_no_stim.get_data(ixs_goods),
                axis=-1))
    
    eigen_values, eigen_vectors = linalg.eig(A, B)
    eigen_values = eigen_values.real
    eigen_vectors = eigen_vectors.real
    ix = np.argsort(eigen_values)[::-1]
    D = eigen_vectors[:, ix].T
    M = linalg.pinv(D)
    n_nulls = _find_n_nulls(A, B, D, M)
    DI = np.ones(M.shape[0])
    DI[:n_nulls] = 0
    DI = np.diag(DI)
    P = M.dot(DI).dot(D)

    if isinstance(obj_stim, RawCLAM):
        obj_stim._data[ixs_goods] = P @ obj_stim._data[ixs_goods]
    else:
        obj_stim._data[:, ixs_goods] = np.array(
            [P @ epoch for epoch in obj_stim._data[:, ixs_goods]])
    
    obj_stim.interpolate_bads(reset_bads=True)


def compute_connectivity(obj, measure='phase_lag_index'):
    
    """
    Compute connectivity between EEG sensors using the phase lag index [1].

    Parameters:
    -----------
    obj : RawCLAM or EpochsCLAM
        The RawCLAM or EpochsCLAM object containing the EEG data.
    measure : str, optional
        The connectivity measure to compute. Currently, only 'phase_lag_index' is supported (default).

    Returns:
    --------
    pandas.DataFrame
        A DataFrame containing the computed connectivity matrices for each trial or epoch.

    Raises:
    -------
    Exception
        If the input data type is not RawCLAM or EpochsCLAM.
        If the connectivity measure is not supported.
        If the data is not filtered into the target frequency range.
        If the data does not contain analytic signals for connectivity computation.
        If there are bad channels that have not been interpolated.
    
    References:
    -----------
    [1] Stam, Cornelis J., Guido Nolte, and Andreas Daffertshofer. "Phase lag index: assessment of 
    functional connectivity from multi channel EEG and MEG with diminished bias from common sources." 
    Human brain mapping 28.11 (2007): 1178-1193.
    """
    
    if not (isinstance(obj, RawCLAM) or isinstance(obj, EpochsCLAM)):
        raise Exception('compute_connectivity can only be applied to RawCLAM or EpochsCLAM objects')

    l_freq = obj.info['highpass']
    h_freq = obj.info['lowpass']
    l_freq_target = obj.l_freq_target
    h_freq_target = obj.h_freq_target
    marker_definition = obj.marker_definition
    participant = obj.participant
    session = obj.session
    design = obj.design
    events = obj.events
    bads = obj.info['bads']
    data = obj.get_data()

    if measure not in ['phase_lag_index']:
        raise Exception('Connectivity measure must be \"phase_lag_index\"')
    if not (l_freq == l_freq_target and h_freq == h_freq_target):
        raise Exception(
            'Data must be filtered into the target frequency range to compute connectivity')
    if not np.iscomplexobj(data):
        raise Exception(
            'Data must contain analytic signal for connectivity computation')
    if not len(bads) == 0:
        raise Exception(
            'Bads must be interpolated before connectivity computation')

    if data.ndim == 2:
        data = data[None, :, :]
    phases = np.angle(data)
    n_chs = phases.shape[1]
    conns = []
    for phase in phases:
        conn = np.zeros((n_chs, n_chs))
        for ix1 in range(n_chs):
            for ix2 in range(ix1 + 1, n_chs):
                conn[ix1, ix2] = misc.pli(phase[ix1], phase[ix2])
        conn += conn.T
        conn += np.diag(np.ones(n_chs))
        conns.append(conn)
    if design == 'trial_wise':
        target_phases = [marker_definition.get(x) for x in events[:, 2]]
    else:
        target_phases = [_get_main_target_phase(marker_definition, events)]
    df_result = pd.DataFrame({'participant': [participant] * len(conns),
                              'session': [session] * len(conns),
                              'design': [design] * len(conns),
                              'target_phase': target_phases,
                              'measure': [measure] * len(conns),
                              'value': conns})
    return df_result